#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#define TEMP_THRESHOLD 75000  // 75℃（1000倍）
#define CLOCK_STEP 50         // クロック調整ステップ（MHz）
#define MIN_CLOCK 800         // 最小クロック（MHz）
#define MAX_CLOCK 1500        // 最大クロック（MHz）
#define INTERVAL 5            // 監視間隔（秒）
#define K_FACTOR 1e-18        // 電力係数
#define R_THERMAL_BASE 0.5    // 基本熱抵抗（℃/W）
#define AMB_TEMP 25000        // 周囲温度25℃（1000倍）
#define THETA_MAX 80000       // 最大許容温度80℃（1000倍）
#define BETA 0.1              // クロック増加のトレードオフ係数
#define FAN_THRESHOLD 70000   // タスク時間計測開始温度70℃（1000倍）
#define FAN_STEP 500          // ファン回転数調整ステップ（RPM）
#define MIN_FAN 1000          // 最小ファン回転数（RPM）
#define MAX_FAN 5000          // 最大ファン回転数（RPM）
#define ETA 0.01              // 冷却係数（W/RPM）
#define C_THERMAL 50          // 熱容量（J/℃）
#define GAMMA 0.5             // ファンvsクロックの判断係数



int read_gpu_temp() {
    FILE *fp = popen("nvidia-smi --query-gpu=temperature.gpu --format=csv,noheader", "r");
    if (!fp) return -1;
    char buffer[16];
    fgets(buffer, sizeof(buffer), fp);
    pclose(fp);
    return atoi(buffer) * 1000;
}

int get_gpu_clock() {
    FILE *fp = popen("nvidia-smi --query-gpu=clocks.gr --format=csv,noheader", "r");
    if (!fp) return -1;
    char buffer[16];
    fgets(buffer, sizeof(buffer), fp);
    pclose(fp);
    return atoi(buffer);
}

void set_gpu_clock(int clock) {
    char cmd[64];
    snprintf(cmd, sizeof(cmd), "nvidia-smi -lgc %d", clock);
    system(cmd);
}

void set_fan_speed(int speed) {
    char cmd[64];
    snprintf(cmd, sizeof(cmd), "nvidia-settings -a [fan:0]/GPUTargetFanSpeed=%d", speed / 50);
    system(cmd);
}

void predict_performance(int clock, double load, double workload, int fan_speed, double *power, double *time, double *temp) {
    double freq = clock * 1e6;
    *power = K_FACTOR * freq * freq * load;
    double r_thermal = R_THERMAL_BASE - 0.0001 * fan_speed;
    *time = workload / (freq * 1000);
    *temp = AMB_TEMP + *power * r_thermal * 1000;
}

int should_increase_clock(double current_time, double new_time, double current_temp, double new_temp) {
    double delta_time = current_time - new_time;
    double delta_temp = new_temp - current_temp;
    double time_ratio = delta_time / current_time;
    double temp_ratio = delta_temp / (THETA_MAX - current_temp);
    return time_ratio > BETA * temp_ratio;
}

int should_increase_fan(double task_time, double delta_theta_fan, double delta_theta_clock, double delta_time_clock) {
    double fan_effect = delta_theta_fan / task_time;
    double clock_effect = delta_theta_clock / delta_time_clock;
    return fan_effect > GAMMA * clock_effect;
}

int main() {
    int current_clock = get_gpu_clock();
    if (current_clock < 0) return 1;

    int current_fan = MIN_FAN;
    double workload = 1e12;
    double load = 0.9;
    double current_power, current_time, current_pred_temp;
    double new_power, new_time, new_pred_temp;
    int task_time_measuring = 0;
    time_t start_time;

    while (1) {
        int temp = read_gpu_temp();
        if (temp < 0) break;

        predict_performance(current_clock, load, workload, current_fan, &current_power, &current_time, &current_pred_temp);

        printf("Temp: %.1f°C, Clock: %d MHz, Fan: %d RPM, Pred Power: %.3f W, Pred Time: %.3f s, Pred Temp: %.1f°C\n",
               temp / 1000.0, current_clock, current_fan, current_power, current_time, current_pred_temp / 1000.0);

        if (temp >= FAN_THRESHOLD && !task_time_measuring) {
            task_time_measuring = 1;
            start_time = time(NULL);
            printf("Started measuring task time at %.1f°C\n", temp / 1000.0);
        }

        if (temp > TEMP_THRESHOLD && current_clock > MIN_CLOCK) {
            int new_clock = current_clock - CLOCK_STEP;
            if (new_clock < MIN_CLOCK) new_clock = MIN_CLOCK;
            set_gpu_clock(new_clock);
            current_clock = new_clock;
        } else if (temp < (TEMP_THRESHOLD - 10000) && current_clock < MAX_CLOCK) {
            int new_clock = current_clock + CLOCK_STEP;
            if (new_clock > MAX_CLOCK) new_clock = MAX_CLOCK;
            predict_performance(new_clock, load, workload, current_fan, &new_power, &new_time, &new_pred_temp);
            if (should_increase_clock(current_time, new_time, temp, new_pred_temp)) {
                set_gpu_clock(new_clock);
                current_clock = new_clock;
            }
        }

        if (task_time_measuring) {
            double task_time = difftime(time(NULL), start_time);
            double delta_theta_fan = (ETA * FAN_STEP * R_THERMAL_BASE) / C_THERMAL;
            double delta_power_clock = K_FACTOR * load * ((current_clock * 1e6) * (current_clock * 1e6) - ((current_clock - CLOCK_STEP) * 1e6) * ((current_clock - CLOCK_STEP) * 1e6));
            double delta_theta_clock = delta_power_clock * R_THERMAL_BASE;
            double delta_time_clock = workload / (((current_clock - CLOCK_STEP) * 1e6) * 1000) - current_time;

            if (temp > TEMP_THRESHOLD && should_increase_fan(task_time, delta_theta_fan, delta_theta_clock, delta_time_clock) && current_fan < MAX_FAN) {
                current_fan += FAN_STEP;
                if (current_fan > MAX_FAN) current_fan = MAX_FAN;
                set_fan_speed(current_fan);
                printf("Increased fan speed to %d RPM\n", current_fan);
            } else if (temp > TEMP_THRESHOLD && current_clock > MIN_CLOCK) {
                int new_clock = current_clock - CLOCK_STEP;
                if (new_clock < MIN_CLOCK) new_clock = MIN_CLOCK;
                set_gpu_clock(new_clock);
                current_clock = new_clock;
                printf("Decreased clock to %d MHz due to temperature\n", current_clock);
            }
        }

        sleep(INTERVAL);
    }

    return 0;
}